/*
 * Copyright 2002 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/param.h>

#include "config.h"

#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/tree.h>
#include <sys/socket.h>
#include <sys/resource.h>

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dnet.h>
#include <syslog.h>

#undef timeout_pending
#undef timeout_initialized

#include <event.h>

#include "honeyd.h"

void
cmd_increase_buf(u_char **pbuf, u_int *psize, u_int maxsize)
{
	u_int size = *psize;
	u_char *p;

	if (size >= maxsize)
		return;
	size *= 2;
	if (size >= maxsize)
		size = maxsize;

	p = realloc(*pbuf, size);
	if (p == NULL)
		return;

	*psize = size;
	*pbuf = p;
}

void
cmd_trigger_read(struct tcp_con *con)
{
 	if (con->pfd == -1 || !con->fdconnected)
		return;
	if (con->psize - con->plen)
		event_add(&con->pread, NULL);
}

void
cmd_trigger_write(struct tcp_con *con)
{
 	if (con->pfd == -1 || !con->fdconnected)
		return;
	if (con->rlen)
		event_add(&con->pwrite, NULL);
}

void
cmd_drain_payload(struct tcp_con *con, u_int len)
{
	if (len == 0 || con->payload == NULL)
		return;

	if (len >= con->plen) {
		con->plen = 0;
		con->poff = 0;
		goto out;
	}

	memmove(con->payload, con->payload + len, con->plen - len);
	con->plen -= len;
	con->poff -= len;
 out:
	cmd_trigger_read(con);
}

int
cmd_add_readbuf(struct tcp_con *con, u_char *dat, u_int datlen)
{
	int space;

	if (con->pfd == -1)
		return (datlen);

	space = con->rsize - con->rlen;
	if (space < datlen) {
		cmd_increase_buf(&con->readbuf, &con->rsize, TCP_MAX_SIZE);
		space = con->rsize - con->rlen;
		if (space < datlen)
			datlen = space;
	}

	memcpy(con->readbuf + con->rlen, dat, datlen);
	con->rlen += datlen;

	cmd_trigger_write(con);

	return (datlen);
}

void
cmd_free(struct tcp_con *con)
{
	event_del(&con->pread);
	event_del(&con->pwrite);
	close(con->pfd);
	con->pfd = -1;
}

void
cmd_read(int fd, short which, void *arg)
{
	struct tcp_con *con = arg;
	int len, space;
	
	space = con->psize - con->plen;
	if (space <= 0)
		return;

	len = read(fd, con->payload + con->plen, space);
	if (len == -1) {
		if (errno == EINTR || errno == EAGAIN)
			goto again;
		cmd_free(con);
		return;
	} else if (len == 0) {
		cmd_free(con);

		tcp_sendfin(con);
		return;
	}

	con->plen += len;
	if (con->plen == con->psize)
		cmd_increase_buf(&con->payload, &con->psize, TCP_MAX_SIZE);

	/* XXX - Trigger write */
	tcp_senddata(con, TH_ACK);

 again:
	cmd_trigger_read(con);
}

void
cmd_write(int fd, short which, void *arg)
{
	struct tcp_con *con = arg;
	int len;
	
	len = write(fd, con->readbuf, con->rlen);
	
	if (len == -1) {
		if (errno == EINTR || errno == EAGAIN)
			goto again;
		cmd_free(con);
		return;
	} else if (len == 0) {
		cmd_free(con);
		return;
	}

	memmove(con->readbuf, con->readbuf + len, con->rlen - len);
	con->rlen -= len;

 again:
	cmd_trigger_write(con);
}

void
cmd_ready_fd(struct tcp_con *con)
{
	event_set(&con->pread, con->pfd, EV_READ, cmd_read, con);
	event_set(&con->pwrite, con->pfd, EV_WRITE, cmd_write, con);
	con->fdconnected = 1;
}

void
cmd_connect_cb(int fd, short which, void *arg)
{
	struct tcp_con *con = arg;
        int error = 0;
        socklen_t errsz = sizeof(error);

	/* Everything is ready */
	cmd_ready_fd(con);

	if (which == EV_TIMEOUT)
		goto out;

        /* Check if the connection completed */
        if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &errsz) == -1) {
                warn("%s: getsockopt for %d", __FUNCTION__, fd);
                goto out;
        }

        if (error) {
                warnx("%s: getsockopt: %s", __FUNCTION__, strerror(error));
		goto out;
	}

	cmd_trigger_read(con);
	cmd_trigger_write(con);
	return;

 out:	
	/* Connection failed, bring this down gracefully */
	cmd_free(con);
	tcp_sendfin(con);
}

struct addrinfo *
cmd_proxy_getinfo(char *address, short port)
{
	struct addrinfo ai, *aitop;
        char strport[NI_MAXSERV];

        memset(&ai, 0, sizeof (ai));
        ai.ai_family = AF_INET;
        ai.ai_socktype = SOCK_STREAM;
        ai.ai_flags = 0;
        snprintf(strport, sizeof (strport), "%d", port);
        if (getaddrinfo(address, strport, &ai, &aitop) != 0) {
                warn("getaddrinfo");
                return (NULL);
        }

	return (aitop);
}

int
cmd_proxy_connect(struct tcp_con *con, struct addrinfo *ai)
{
	char ntop[NI_MAXHOST], strport[NI_MAXSERV];
	char *host = ntop, *port = strport;
	struct timeval tv = {10, 0};
        int fd;
        
        fd = socket(AF_INET, SOCK_STREAM, 0);
        if (fd == -1) {
                warn("socket");
                return (-1);
        }

        if (fcntl(fd, F_SETFL, O_NONBLOCK) == -1)
                warn("fcntl(O_NONBLOCK)");

        if (fcntl(fd, F_SETFD, 1) == -1)
                warn("fcntl(F_SETFD)");

	con->pfd = fd;
        if (connect(fd, ai->ai_addr, ai->ai_addrlen) == 0) {
		cmd_connect_cb(fd, EV_WRITE, con);
		return (0);
	}

	if (errno != EINPROGRESS) {
		warn("connect");
		con->pfd = -1;
		close(fd);
		return (-1);
	}

	event_set(&con->pwrite, fd, EV_WRITE, cmd_connect_cb, con);
	event_add(&con->pwrite, &tv);

	if (getnameinfo(ai->ai_addr, ai->ai_addrlen,
		ntop, sizeof(ntop), strport, sizeof(strport),
		NI_NUMERICHOST|NI_NUMERICSERV) != 0) {

		host = "<hosterror>";
		port = "<porterror>";
	}
	syslog(LOG_INFO, "Connection established: %s -> proxy to %s:%s",
	    honeyd_contoa(con), host, port);

	return (0);
}

int
cmd_fork(struct tcp_con *con, char *cmd, char **argv)
{
	extern int honeyd_nchildren;
	int pair[2];
	sigset_t sigmask;

	if (socketpair(AF_UNIX, SOCK_STREAM, 0, pair) == -1)
		return (-1);

	/* Block SIGCHLD */
	sigemptyset(&sigmask);
	sigaddset(&sigmask, SIGCHLD);
	if (sigprocmask(SIG_BLOCK, &sigmask, NULL) == -1) {
		warn("sigprocmask");
		goto fork_err;
	}

	con->pid = fork();
	if (con->pid == -1) {
		warn("fork");
		goto unmask_err;
	}

	if (con->pid == 0) {
		struct template *tmpl = con->tmpl;
		uid_t uid = 32767;
		gid_t gid = 32767;
		struct rlimit rl;

		if (tmpl->uid)
			uid = tmpl->uid;
		if (tmpl->gid)
			gid = tmpl->gid;

		/* Lower privileges */
#ifdef HAVE_SETGROUPS
		setgroups(1, &gid);
#endif
		setegid(gid);
		setgid(gid);
		seteuid(uid);
		setuid(uid);

		/* Raising file descriptor limits */
		rl.rlim_cur = rl.rlim_max = 24;
		if (setrlimit(RLIMIT_NOFILE, &rl) == -1)
			err(1, "setrlimit");

		/* Child */
		close(pair[0]);
		if (dup2(pair[1], fileno(stdout)) == -1)
			err(1, __FUNCTION__": dup2");
		if (dup2(pair[1], fileno(stdin)) == -1)
			err(1, __FUNCTION__": dup2");

		close(pair[1]);

		if (execv(cmd, argv) == -1)
			err(1, __FUNCTION__": execv(%s)", cmd);

		/* NOT REACHED */
	}

	close(pair[1]);
	con->pfd = pair[0];
	if (fcntl(con->pfd, F_SETFD, 1) == -1)
		warn("fcntl(F_SETFD)");
	if (fcntl(con->pfd, F_SETFL, O_NONBLOCK) == -1)
		warn("fcntl(F_SETFL)");

	cmd_ready_fd(con);

	event_add(&con->pread, NULL);

	honeyd_nchildren++;

	/* Install old signal handler */
	if (sigprocmask(SIG_UNBLOCK, &sigmask, NULL) == -1) {
		warn("sigprocmask");
		goto fork_err;
	}
	return (0);

	/* Error cleanup */
 unmask_err:
	/* Install old signal handler */
	if (sigprocmask(SIG_UNBLOCK, &sigmask, NULL) == -1)
		warn("sigprocmask");

 fork_err:
	close(pair[0]);
	close(pair[1]);
	con->pfd = -1;

	return (-1);
}
